///////////////////////////////////////////////////////////////////////
// Copyright (c) 2017 VMware, Inc. All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0
///////////////////////////////////////////////////////////////////////

package cmd

import (
	"io"
	"time"

	"golang.org/x/net/context"

	"github.com/olekukonko/tablewriter"
	"github.com/spf13/cobra"

	"github.com/vmware/dispatch/pkg/api/v1"
	"github.com/vmware/dispatch/pkg/client"
	"github.com/vmware/dispatch/pkg/dispatchcli/i18n"
)

var (
	getBaseImagesLong = i18n.T(`Get base images.`)

	// TODO: add examples
	getBaseImagesExample = i18n.T(``)
)

// NewCmdGetBaseImage creates command responsible for getting base images.
func NewCmdGetBaseImage(out io.Writer, errOut io.Writer) *cobra.Command {
	cmd := &cobra.Command{
		Use:     "base-image [IMAGE_NAME]",
		Short:   i18n.T("Get base images"),
		Long:    getBaseImagesLong,
		Example: getBaseImagesExample,
		Args:    cobra.MaximumNArgs(1),
		Aliases: []string{"base-images", "baseimage", "baseimages"},
		Run: func(cmd *cobra.Command, args []string) {
			var err error
			c := baseImagesClient()
			if len(args) > 0 {
				err = getBaseImage(out, errOut, cmd, args, c)
			} else {
				err = getBaseImages(out, errOut, cmd, c)
			}
			CheckErr(err)
		},
	}
	return cmd
}

func getBaseImage(out, errOut io.Writer, cmd *cobra.Command, args []string, c client.BaseImagesClient) error {
	baseImageName := args[0]
	resp, err := c.GetBaseImage(context.TODO(), dispatchConfig.Organization, baseImageName)
	if err != nil {
		return err
	}
	return formatBaseImageOutput(out, false, []v1.BaseImage{*resp})
}

func getBaseImages(out, errOut io.Writer, cmd *cobra.Command, c client.BaseImagesClient) error {
	resp, err := c.ListBaseImages(context.TODO(), dispatchConfig.Organization)
	if err != nil {
		return err
	}
	return formatBaseImageOutput(out, true, resp)
}

func formatBaseImageOutput(out io.Writer, list bool, images []v1.BaseImage) error {
	if w, err := formatOutput(out, list, images); w {
		return err
	}
	table := tablewriter.NewWriter(out)
	table.SetHeader([]string{"Name", "Language", "URL", "Status", "Created Date"})
	table.SetBorders(tablewriter.Border{Left: false, Top: false, Right: false, Bottom: false})
	table.SetCenterSeparator("")
	for _, image := range images {
		table.Append([]string{image.Name, *image.Language, *image.ImageURL, string(image.Status), time.Unix(image.CreatedTime, 0).Local().Format(time.UnixDate)})
	}
	table.Render()
	return nil
}
